import time
from functools import lru_cache

from hhelper.ui.print_greeting import print_greeting


@lru_cache(maxsize=1)
def format_options(options):
    formatted_options = []

    for index, option in enumerate(options):
        formatted_options.append(f"{index + 1}. {option}")

    max_len = max(len(o) for o in formatted_options) + 1

    formatted_options = [o.ljust(max_len) for o in formatted_options]

    return formatted_options


def display_menu(options, term, len_options, selected_index, is_jump=False):
    print(term.home + term.move_y(term.height // 2 - len_options // 2 - 8))
    print_greeting(term)

    highlight_color_funct = (
        term.bold_white_on_yellow if is_jump else term.bold_white_on_green
    )

    menu_options = format_options(options)

    for index, option in enumerate(menu_options):
        if index == selected_index:
            print(term.center(highlight_color_funct(option)))
        else:
            print(term.center(term.white(option)))


def menu(options, term):
    print(term.clear + term.home)

    selected_index = 0

    len_options = len(options)
    end = len_options - 1

    num_key_option = {str(i) for i in range(1, len_options + 1)}

    with term.cbreak(), term.hidden_cursor():
        display_menu(options, term, len_options, selected_index)
        while True:
            key = term.inkey()

            if key in num_key_option:
                selected_index = int(key) - 1
                display_menu(options, term, len_options, selected_index, is_jump=True)
                time.sleep(0.2)
                break

            elif key.name == "KEY_ENTER":
                display_menu(options, term, len_options, selected_index, is_jump=True)
                time.sleep(0.2)
                break
            elif key.name == "KEY_UP":
                if selected_index == 0:
                    selected_index = end

                else:
                    selected_index -= 1
                display_menu(options, term, len_options, selected_index)
            elif key.name == "KEY_DOWN":
                if selected_index == end:
                    selected_index = 0
                else:
                    selected_index += 1
                display_menu(options, term, len_options, selected_index)
            else:
                pass

    return options[selected_index]



import subprocess

from hhelper.helpers.return_status import STATUS
from hhelper.ui.display import press_key_to_continue


def generate_recurring_tx(ledger_path, recurring_tx_path, term):
    while True:
        print(term.clear + term.home)
        print(term.move_y(term.height))
        period_expression = input(
            term.green("Input a period expression, or q to quit, or ? for help: ")
        )

        if period_expression.lower() in {"q", "quit"}:
            return STATUS.NOWAIT

        elif period_expression == "?":
            print(term.clear + term.home + term.move_y(term.height))
            print(
                """Valid period expressions include: "..", "apr", "april", "aug", "august", "bimonthly", "biweekly", "daily", "dec", "december", "every", "feb", "february", "fortnightly", "from", "in", "jan", "january", "jul", "july", "jun", "june", "last", "mar", "march", "may", "monthly", "next", "nov", "november", "oct", "october", "quarterly", "sep", "september", "this", "to", "today", "tomorrow", "until", "weekly", "yearly", "yesterday", '+', '-', 'Q', 'q', digit, integer, or year."""
            )
            press_key_to_continue(term)

            continue

        else:
            recurring_tx_result = subprocess.run(
                [
                    "hledger",
                    f"--file={str(recurring_tx_path)}",
                    "print",
                    "--forecast",
                    "-p",
                    period_expression,
                ],
                capture_output=True,
                text=True,
            )

            if recurring_tx_result.stderr != "":
                print(recurring_tx_result.stderr)

                press_key_to_continue(term)

                continue

            else:
                recur_tx = recurring_tx_result.stdout
                print(term.clear + term.home)
                print(term.move_y(term.height))

                print(recur_tx)

                decision = input(
                    term.green("Append this to jorunal? (y/N/q): ")
                ).lower()

                if decision in {"y", "yes"}:
                    with open(ledger_path, "a") as f:
                        f.write(recur_tx)

                elif decision in {"", "n", "no", "q", "quit"}:
                    continue

                else:
                    raise ValueError



import datetime
import re
from collections import OrderedDict
from enum import Enum
from functools import cache

from hhelper.helpers.check_valid_journal import check_valid_journal
from hhelper.helpers.return_status import STATUS
from hhelper.ui.display import clear_screen_move_to_bottom, press_key_to_continue

line_type = Enum(
    "Type",
    ["CLEARED", "UNCLEARED_HEAD", "UNCLEARED_BODY", "GENERATED_COMMENTS"],
)
search_string_type = Enum("Type", ["ALL", "QUIT"])
tx_decision_type = Enum(
    "Type",
    ["YES_CLEAR", "YES_CLEAR_ALL", "DONT_CLEAR", "VIEW_REST", "QUIT", "REGEX", "HELP"],
)


def get_tx_decision(prefix, tx, term):
    while True:
        print(prefix)
        print(tx, end="", flush=True)
        try:
            user_input = input(
                term.green("Clear Transaction (y/n/q/a/v/r/h): ")
            ).lower()

            if user_input in {"", "y", "yes"}:
                return tx_decision_type.YES_CLEAR
            elif user_input in {"n", "no"}:
                return tx_decision_type.DONT_CLEAR
            elif user_input in {"q", "quit"}:
                return tx_decision_type.QUIT
            elif user_input in {"a", "all"}:
                return tx_decision_type.YES_CLEAR_ALL
            elif user_input in {"v", "view"}:
                return tx_decision_type.VIEW_REST
            elif user_input in {"r", "regex"}:
                return tx_decision_type.REGEX
            elif user_input in {"h", "help"}:
                return tx_decision_type.HELP

        except (KeyboardInterrupt, EOFError):
            print("Interrupted")
            print("Bye!")
            exit()


def get_regex_search_string(term):
    try:
        search_string = input(
            term.green(
                'Regex query for transaction (leave blank for no filter, "q" or "quit" for menu): '
            )
        )

        if search_string.lower() in {"q", "quit"}:
            return search_string_type.QUIT
        elif search_string == "":
            return search_string_type.ALL
        else:
            return search_string
    except (KeyboardInterrupt, EOFError):
        print("Interrupted")
        print("Bye!")
        exit()


@cache
def is_transaction_header(text):
    match = (
        re.match(r"((?P<year>\d{4})-)?(?P<month>\d{1,2})-(?P<day>\d{1,2}) .*", text)
        or re.match(
            r"((?P<year>\d{4})\.)?(?P<month>\d{1,2})\.(?P<day>\d{1,2}) .*", text
        )
        or re.match(r"((?P<year>\d{4})/)?(?P<month>\d{1,2})/(?P<day>\d{1,2}) .*", text)
    )

    if match:
        try:
            month = int(match.group("month"))
            day = int(match.group("day"))

            if match.group("year"):
                year = int(match.group("year"))
                datetime.date(year=year, month=month, day=day)
            else:
                # If year is not provided, just validate month and day
                datetime.date(
                    year=2000, month=month, day=day
                )  # Use a leap year to allow Feb 29

            return True
        except (ValueError, TypeError):
            return False
    else:
        return False


@cache
def is_transaction_header_cleared(text):
    if is_transaction_header(text):
        match = (
            re.match(
                r"((?P<year>\d{4})-)?(?P<month>\d{1,2})-(?P<day>\d{1,2}) \* ", text
            )
            or re.match(
                r"((?P<year>\d{4})\.)?(?P<month>\d{1,2})\.(?P<day>\d{1,2}) \* ", text
            )
            or re.match(
                r"((?P<year>\d{4})/)?(?P<month>\d{1,2})/(?P<day>\d{1,2}) \* ", text
            )
        )

        return match

    else:
        return False


def update_line_status(lines, start_line):
    line_status = {}
    uncleared_tx = {}
    uncleared_tx_text = {}
    num_unclear = 0

    for line_number, line in lines.items():
        if line_number < start_line:
            pass

        else:
            if is_transaction_header(line) and not is_transaction_header_cleared(line):
                line_status[line_number] = line_type.UNCLEARED_HEAD

                uncleared_tx[line_number] = [line_type.UNCLEARED_HEAD]
                uncleared_tx_text[line_number] = [line]

                current_unclear_head = line_number

                num_unclear += 1

            elif (
                line_number >= start_line + 1
                and line_status[line_number - 1] == line_type.UNCLEARED_HEAD
                and line.strip().startswith("; generated-transaction:")
            ):
                line_status[line_number] = line_type.GENERATED_COMMENTS

                uncleared_tx[current_unclear_head].append(line_type.GENERATED_COMMENTS)
                uncleared_tx_text[current_unclear_head].append(line)

            elif (
                line_number >= start_line + 1
                and line_status[line_number - 1]
                in {line_type.UNCLEARED_HEAD, line_type.UNCLEARED_BODY}
                and re.match("\s+\w+", line)
            ):
                line_status[line_number] = line_type.UNCLEARED_BODY

                uncleared_tx[current_unclear_head].append(line_type.UNCLEARED_BODY)
                uncleared_tx_text[current_unclear_head].append(line)

            elif (
                line_number >= start_line + 2
                and line_status[line_number - 2] == line_type.UNCLEARED_HEAD
                and line_status[line_number - 1] == line_type.GENERATED_COMMENTS
            ):
                line_status[line_number] = line_type.UNCLEARED_BODY

                uncleared_tx[current_unclear_head].append(line_type.UNCLEARED_BODY)
                uncleared_tx_text[current_unclear_head].append(line)
            else:
                line_status[line_number] = line_type.CLEARED

    uncleared_tx_text = {k: "".join(v) for k, v in uncleared_tx_text.items()}

    return uncleared_tx, uncleared_tx_text, num_unclear


def print_help_string():
    print(
        "y/yes: clear current transaction",
        "n/no: don't clear current transaction",
        "q/quit: quit to main menu",
        "a/all: clear all the remaining transaction in current query",
        "v/view: view remaining transaction in current query",
        "r/regex: enter new regex query",
        "h/help: print this help",
        "",
        "If any, modifications will be written out to file upon each selection.",
        sep="\n",
    )


def clear_tx(ledger_path, term):
    with open(ledger_path, "r") as f:
        lines = f.readlines()

    check_valid_journal("".join(lines))

    lines = OrderedDict([(index, line) for index, line in enumerate(lines, start=1)])

    starting_line = 1

    unclear_query_pattern = "|".join(
        [
            r"((\d{4}-)?\d{1,2}-\d{1,2} )(! )?",
            r"((\d{4}/)?\d{1,2}/\d{1,2} )(! )?",
            r"((\d{4}\.)?\d{1,2}\.\d{1,2} )(! )?",
        ]
    )

    unclear_query_pattern = re.compile(f"^({unclear_query_pattern})")

    while True:
        clear_screen_move_to_bottom(term)
        uncleared_tx, uncleared_tx_text, uncleared_count = update_line_status(
            lines, starting_line
        )

        if uncleared_count == 0:
            print("All cleared. Bye!")
            return STATUS.WAIT

        print(term.yellow(f"{uncleared_count} uncleared transaction left."))
        starting_line = min(uncleared_tx.keys())

        search_string = get_regex_search_string(term)
        clear_screen_move_to_bottom(term)

        if search_string == search_string_type.QUIT:
            return STATUS.NOWAIT
        elif search_string == search_string_type.ALL:
            pass
        elif search_string != search_string_type.ALL:
            uncleared_tx_text = {
                k: v
                for k, v in uncleared_tx_text.items()
                if re.search(search_string, v, flags=re.I)
            }

            uncleared_tx = {
                k: v for k, v in uncleared_tx.items() if k in uncleared_tx_text
            }
        else:
            raise ValueError

        keys = list(uncleared_tx_text.keys())
        total_num = len(keys)
        max_index = total_num - 1

        index = 0
        clear_all_flag = False
        while index <= max_index:
            k = keys[index]
            v = uncleared_tx_text[k]

            index += 1

            if clear_all_flag:
                decision = tx_decision_type.YES_CLEAR_ALL
            else:
                decision = get_tx_decision(f"[{index}/{total_num}]", v, term)

            if decision == tx_decision_type.HELP:
                clear_screen_move_to_bottom(term)
                print_help_string()
                press_key_to_continue(term)
                clear_screen_move_to_bottom(term)

                index -= 1
                continue

            elif decision == tx_decision_type.REGEX:
                clear_screen_move_to_bottom(term)
                break

            elif decision == tx_decision_type.QUIT:
                return STATUS.NOWAIT

            elif decision == tx_decision_type.DONT_CLEAR:
                clear_screen_move_to_bottom(term)

            elif decision == tx_decision_type.VIEW_REST:
                remaining_items = [
                    value for key, value in uncleared_tx_text.items() if key >= k
                ]

                num_remaining = len(remaining_items)

                clear_screen_move_to_bottom(term)
                for i, item in enumerate(remaining_items, start=1):
                    print(f"[{i}/{num_remaining}]")
                    print(item)

                for _ in range(2):
                    print("*" * term.width)

                print()

                index -= 1

            elif decision in {
                tx_decision_type.YES_CLEAR,
                tx_decision_type.YES_CLEAR_ALL,
            }:
                if decision == tx_decision_type.YES_CLEAR_ALL:
                    clear_all_flag = True

                lines[k] = unclear_query_pattern.sub(r"\2* ", lines[k])

                if uncleared_tx[k][1] == line_type.GENERATED_COMMENTS:
                    del lines[k + 1]

                with open(ledger_path, "w") as f:
                    for line in lines.values():
                        f.write(line)
                clear_screen_move_to_bottom(term)
            else:
                raise NotImplementedError

        lines = OrderedDict(
            [(index, lines[k]) for index, k in enumerate(lines.keys(), start=1)]
        )



def press_key_to_continue(term):
    with term.cbreak(), term.hidden_cursor():
        print(term.bold_white_on_green("Press any key to continue"), end="", flush=True)

        term.inkey()


def clear_screen_move_to_bottom(term):
    print(term.clear() + term.home() + term.move_y(term.height))



from pathlib import Path

import tomllib
from blessed import Terminal
from hhelper.helpers.backup import backup_file
from hhelper.helpers.options import (
    AvailableHelpers,
    get_main_menu_options,
    get_selected_option,
)
from hhelper.helpers.return_status import STATUS
from hhelper.ui.display import press_key_to_continue
from hhelper.ui.menu import menu


def main():
    config_path = Path().home() / ".config" / "hhelper" / "config.toml"

    if config_path.is_file():
        with open(config_path, "rb") as f:
            config = tomllib.load(f)

    else:
        raise FileNotFoundError(
            (
                f"The config file is not found. Please create a config file at {config_path}.\n"
                f"You can find an example config file at https://github.com/plwg/hledger_helper"
            )
        )

    paths = config["paths"]

    directory = Path(paths["directory"]).expanduser()

    ledger_path = directory / paths["ledger_file"]
    price_path = directory / paths["price_file"]
    header_path = directory / paths["header_file"]
    recurring_tx_path = directory / paths["recurring_tx_file"]

    term = Terminal()

    while True:
        selection = menu(get_main_menu_options(), term)

        name, helper = get_selected_option(selection, term)

        print(term.clear + term.home)
        print(term.move_y(term.height))

        if name == AvailableHelpers.FETCH_PRICE:
            commodity_pairs = config["commodities"]["commodity_pairs"]

            backup_file(price_path)
            status = helper(price_path, commodity_pairs, term)

        elif name == AvailableHelpers.MARK_CLEAR:
            backup_file(ledger_path)
            status = helper(ledger_path, term)

        elif name == AvailableHelpers.CLEAN_UP:
            bak_up_path = backup_file(ledger_path)
            backup_file(header_path)
            status = helper(ledger_path, header_path, bak_up_path, term)

        elif name == AvailableHelpers.GEN_RECUR:
            backup_file(ledger_path)

            status = helper(ledger_path, recurring_tx_path, term)

        else:
            raise NotImplementedError

        if status == STATUS.WAIT:
            press_key_to_continue(term)

        elif status == STATUS.NOWAIT:
            pass

        else:
            raise ValueError


if __name__ == "__main__":
    main()



import re
from functools import cache


@cache
def get_separator_length(line):
    matches = re.findall(r" {2,}", line.lstrip())
    if matches:
        return len(matches[0])
    else:
        return -1


@cache
def expand_separator(line, new_length):
    left_most_white_space = re.match(r"^\s*", line).group()

    before_separator, after_separator = re.split(r" {2,}", line.lstrip(), maxsplit=1)

    new_separator = " " * new_length
    return "".join(
        [left_most_white_space, f"{before_separator}{new_separator}{after_separator}"]
    )


def align_amounts(text):
    lines = text.split("\n")
    anchor_locations = []

    for line in lines:
        if re.search(r"\d\.\d", line):
            anchor_locations.append(re.search(r"\d\.\d", line).span()[0])

        else:
            anchor_locations.append(-1)

    max_anchor_location = max(anchor_locations)

    for index, line in enumerate(lines):
        anchor_loc = anchor_locations[index]

        if anchor_loc != -1:
            diff = max_anchor_location - anchor_loc

            lines[index] = expand_separator(line, get_separator_length(line) + diff)

        else:
            pass

    return "\n".join(lines)



from collections import namedtuple
from enum import Enum, auto

from hhelper.helpers.clean_up_journal import clean_up_journal
from hhelper.helpers.clear_tx import clear_tx
from hhelper.helpers.fetch_price import fetch_price
from hhelper.helpers.generate_recurring_tx import generate_recurring_tx

Helper = namedtuple("Helper", ["name", "function"])


class AvailableHelpers(Enum):
    MARK_CLEAR = auto()
    CLEAN_UP = auto()
    FETCH_PRICE = auto()
    GEN_RECUR = auto()


_helpers = {
    AvailableHelpers.MARK_CLEAR: Helper("Mark Transactions as Cleared", clear_tx),
    AvailableHelpers.CLEAN_UP: Helper("Clean Up Journal", clean_up_journal),
    AvailableHelpers.FETCH_PRICE: Helper("Fetch Prices", fetch_price),
    AvailableHelpers.GEN_RECUR: Helper(
        "Generate Recurring Transactions", generate_recurring_tx
    ),
}


def get_main_menu_options():
    menu_options = sorted(helper.name for helper in _helpers.values())
    menu_options.append("Exit")
    return tuple(menu_options)


def get_selected_option(option, term):
    if option == "Exit":
        print(term.clear + term.home)

        exit()

    for k, v in _helpers.items():
        if v.name == option:
            return k, v.function

    raise ValueError(f"Invalid option: {option}")



#!./.venv/bin/python
import re
from collections import namedtuple
from datetime import datetime as dt
from datetime import timedelta as timedelta

import yfinance as yf

from .return_status import STATUS


def fetch_hist_price(name, start_date):
    price_history = yf.download(
        name, start=start_date, interval="1d", progress=False, multi_level_index=False
    )

    return price_history


def parse_hledger_format(price_history, commodity1, commodity2, append_space):
    prices = []

    for index, row in price_history.iterrows():
        prices.append(
            f"P {index.date()} {commodity1} {commodity2}{' ' if append_space else ''}{round(row['Close'], 2)}\n"
        )

    return prices


def fetch_price(price_file_path, commodity_pairs, term):
    with open(price_file_path, "r") as file_object:
        lines = file_object.readlines()

    date_pat = re.compile(r"\d\d\d\d-\d\d-\d\d")

    latest_date = max(date_pat.search(line).group(0) for line in lines)

    latest_date = dt.strptime(latest_date, "%Y-%m-%d")
    start_date = latest_date - timedelta(days=30)
    start_date_str = str(start_date)[:10]

    daily_price = []

    CommodityPair = namedtuple("CommodityPair", commodity_pairs[0])

    commodity_pairs = [CommodityPair(**cp) for cp in commodity_pairs]

    for pair in commodity_pairs:
        daily_price.extend(
            parse_hledger_format(
                fetch_hist_price(pair.symbol, start_date),
                pair.base_currency,
                pair.quote_currency,
                pair.is_append_space,
            )
        )

    latest_date = max(date_pat.search(line).group(0) for line in daily_price)

    print(term.clear + term.home)
    print("".join(daily_price))
    print(f"Fetched {len(daily_price)} postings from {start_date_str} to {latest_date}")

    descision = input(term.green("Write to file? (Y/n/q): ")).lower()

    if descision in {"", "y", "yes"}:
        pass

    elif descision in {"n", "no", "q", "quit"}:
        return STATUS.NOWAIT

    else:
        raise ValueError

    for line in lines:
        if date_pat.search(line).group(0) < start_date_str:
            daily_price.append(line)

    daily_price.sort()

    with open(price_file_path, "w") as file_object:
        file_object.writelines(daily_price)

    print(f"Prices successfully written to {price_file_path}")

    return STATUS.WAIT



import subprocess
from functools import cache


@cache
def check_valid_journal(text):
    result = subprocess.run(
        ["hledger", "check", "-f", "-"], input=text, text=True, capture_output=True
    )
    err = result.stderr

    if err != "":
        print(err)
        exit()



#!/usr/bin/env python3
import subprocess

from hhelper.helpers.align_posting import align_amounts
from hhelper.helpers.check_valid_journal import check_valid_journal
from hhelper.helpers.return_status import STATUS


def clean_up_journal(ledger_path, header_path, backup_ledger_path, term):
    with term.hidden_cursor():
        warning_message = (
            "Relying on hledger print, cleaning up has the following behaviors:",
            "- All amounts are shown explicity. Balance assignments are converted into balance assertions.",
            "- Amounts are renormalized to commodity display style.",
            "- All decimal places are shown.",
            "- Directives and inter-transaction comments are lost.",
            "- Valuation affects posting amounts but not balance assertion or balance assignment amounts",
            "  potentially causing those to fail.",
            "- Auto postings can generate postings with too many missing amounts.",
            "- Account aliases can generate bad account names.",
            "- Amounts across transactions are aligned based on decimal point location.",
            "",
            "The cleaner will check that the resulting ledger is a valid journal before writing back to the file.",
            f"The unmodified journal can be found at {backup_ledger_path}",
        )

        max_len = max(len(m) for m in warning_message)
        warning_message = [m.ljust(max_len) for m in warning_message]

        print(term.home + term.move_y(term.height // 2 - len(warning_message) // 2))

        for msg in warning_message:
            print(term.center(term.bold(msg)))

        print(term.move_y(term.height))

        descision = input(term.green("Continue? (Y/n/q): ")).lower()

        if descision in {"", "y", "yes"}:
            pass

        elif descision in {"n", "no", "q", "quit"}:
            return STATUS.NOWAIT

        else:
            raise ValueError

    print(term.clear + term.move_y(term.height))

    # Generate the ledger content
    sorted_ledger = subprocess.run(
        ["hledger", "print", "-x", "-f", str(ledger_path)],
        capture_output=True,
        text=True,
    ).stdout

    # NOTE: There is no need to account for header here since it is not in the hledger print's output
    sorted_ledger = align_amounts(sorted_ledger)

    # Read the header file
    with open(header_path, "r") as header_file:
        header_content = header_file.read()
    print(term.bold_white(f"Read header file from {header_path}"))

    # Check the result ledger
    sorted_ledger = f"{header_content}\n\n\n{sorted_ledger}"

    check_valid_journal(sorted_ledger)

    # Write the sorted ledger to the file
    with open(ledger_path, "w") as ledger_file:
        ledger_file.write(sorted_ledger)

    print(term.bold_white(f"Write sorted ledger to {ledger_path}"))

    return STATUS.WAIT
